#include "break.h"
#include "log_calendar.h"
#include <e32base.h>
#include "log_base_impl.h"
#include "app_context.h"
#include "symbian_auto_ptr.h"
#include "csd_calendar.h"

#include <agmcallb.h>
#include <agmmiter.h>
#include "timeout.h"

TInt AgnNotify(TAny* aPtr);


class CLogCalendarImpl : public CLogCalendar, 
	public MAgnProgressCallBack, public MAgnModelStateCallBack, public MTimeOut {
private:
	CLogCalendarImpl(MApp_context& Context);
	void ConstructL();
	~CLogCalendarImpl();

	void ReadEntriesL(TBool aForce=EFalse);
	void ListCurrentL();

	// MAgnProgressCallBack
	void Completed(TInt aError = KErrNone);
	void Progress(TInt aPercentageCompleted) { }
	// MAgnModelStateCallBack
	void StateCallBack(CAgnEntryModel::TState aState); 
	// MTimeOut
	void expired(CBase* Source);

	enum TState { EInitializing, EOpening, EOpened };
	TState iState;
	CAgnEntryModel::TState iModelState;

	RAgendaServ* iAgnServ;
	CAgnModel*	iModel;
	CAgnDayDateTimeInstanceList*	iDayList;
	TDateTime	iDayListDate;
	CTimeOut	*iTimeOut;

	TBBCalendar	iValue;

	friend class CLogCalendar;
	friend class auto_ptr<CLogCalendarImpl>;
	friend TInt AgnNotify(TAny* aPtr);
};

EXPORT_C CLogCalendar* CLogCalendar::NewL(class MApp_context& Context)
{
	CALLSTACKITEM2_N(_CL("CLogCalendar"), _CL("NewL"), &Context);
	auto_ptr<CLogCalendarImpl> ret(new (ELeave) CLogCalendarImpl(Context));
	ret->ConstructL();
	return ret.release();
}

CLogCalendar::CLogCalendar(MApp_context& Context) : Mlog_base_impl(Context, KCalendar, KCalendarTuple, 24*60*60) { }

CLogCalendarImpl::CLogCalendarImpl(MApp_context& Context) : CLogCalendar(Context) { }

TInt AgnNotify(TAny* aPtr)
{
	CALLSTACKITEM_N(_CL("CLogCalendarImpl"), _CL("AgnNotify"));
	CLogCalendarImpl* log=(CLogCalendarImpl*)aPtr;
	TInt err;
	CC_TRAP(err, log->ReadEntriesL(ETrue));
	if (err!=KErrNone) return err;
	CC_TRAP(err, log->ListCurrentL());
	if (err!=KErrNone) return err;
	return KErrNone;
}

void CLogCalendarImpl::ConstructL()
{
	CALLSTACKITEM_N(_CL("CLogCalendarImpl"), _CL("ConstructL"));

	iDayListDate=TTime(0).DateTime();
	Mlog_base_impl::ConstructL();

	iTimeOut=CTimeOut::NewL(*this);

	TTimeIntervalMinutes KDefaultTimeForEvents(9 * 60); // 9.00 pm
	TTimeIntervalMinutes KDefaultTimeForAnnivs(13 * 60); // 1.00 am
	TTimeIntervalMinutes KDefaultTimeForDayNote(13 * 60); // 1.00 am

	iAgnServ = RAgendaServ::NewL();
	iAgnServ->Connect();

	iAgnServ->StartNotifierL( TCallBack(AgnNotify, this) );

	TTime now; now.HomeTime();
	iDayList=CAgnDayDateTimeInstanceList::NewL(now);

	iModel=CAgnModel::NewL(this);
	iModel->SetServer(iAgnServ);
	iModel->SetMode(CAgnEntryModel::EClient); 

	iState = EOpening;
	iModel->OpenL(_L("C:\\system\\data\\calendar"),
		KDefaultTimeForEvents, KDefaultTimeForAnnivs, KDefaultTimeForDayNote,
		this);


	iEvent.iData()=&iValue; iEvent.iData.SetOwnsValue(EFalse);
	post_new_value(&iValue);
}

void CLogCalendarImpl::Completed(TInt aError)
{
	CALLSTACKITEM_N(_CL("CLogCalendarImpl"), _CL("Completed"));
	User::LeaveIfError(aError);
	if (iState==EOpening) iState=EOpened;
	if (iModelState==CAgnEntryModel::EOk) {
		ReadEntriesL();
		ListCurrentL();
	}
}

void CLogCalendarImpl::StateCallBack(CAgnEntryModel::TState aState)
{
	iModelState=aState;
	if (iModelState==CAgnEntryModel::EOk) {
		ReadEntriesL();
		ListCurrentL();
	}
}

void MakeCalendarEvent(TBBCalendarEvent& aInto, CAgnDayDateTimeInstanceList* aFromList, CAgnModel* aModel)
{
	CALLSTACKITEM_N(_CL("User"), _CL("LeaveIfError"));
	TInt count=aFromList->Count();
	TTime min(0), max(0);
	aInto.iStartTime() = min;
	aInto.iEndTime() = max;

	for (TInt i=0; i<count; i++) {
		TAgnInstanceDateTimeId inst=(*aFromList)[i];
		auto_ptr<CAgnEntry> entry(aModel->FetchInstanceL(inst));
		TInt space_left=aInto.iDescription().MaxLength()-aInto.iDescription().Length();
		if ( aInto.iDescription().Length() > 0  && space_left>=2) {
			aInto.iDescription().Append(_L("; "));
			space_left-=2;
		}
		TBuf<30> desc;
		entry->RichTextL()->Extract(desc,0,30);
		TInt to_append_len=desc.Length();
		if (to_append_len > space_left) to_append_len=space_left;
		aInto.iDescription().Append ( desc.Left(to_append_len ) );

		if ( min==TTime(0) || entry->InstanceStartDate() < min ) min=entry->InstanceStartDate();
		if ( max==TTime(0) || entry->InstanceEndDate() > max ) max=entry->InstanceEndDate();
	}
	aInto.iStartTime() = min;
	aInto.iEndTime() = max;
}

void CLogCalendarImpl::ListCurrentL()
{
	CALLSTACKITEM_N(_CL("CLogCalendarImpl"), _CL("ListCurrentL"));

	iTimeOut->Wait( 10*60 );

	if (iState!=EOpened) return;
	if (iModelState!=CAgnEntryModel::EOk) return;

	TTime now; now.HomeTime();
	TDateTime midnight=now.DateTime(); midnight.SetHour(0); 
	midnight.SetMinute(0); midnight.SetSecond(0);
	TTimeIntervalMinutes now_min_int;
	User::LeaveIfError(now.MinutesFrom(midnight, now_min_int));
	TInt now_min=now_min_int.Int();

	TInt count=iDayList->Count();

	auto_ptr<CAgnDayDateTimeInstanceList> prev(CAgnDayDateTimeInstanceList::NewL(now));
	auto_ptr<CAgnDayDateTimeInstanceList> current(CAgnDayDateTimeInstanceList::NewL(now));
	auto_ptr<CAgnDayDateTimeInstanceList> next(CAgnDayDateTimeInstanceList::NewL(now));

	TBBCalendar current_cal;

	for (TInt i=0; i<count; i++) {
		TAgnInstanceDateTimeId inst=(*iDayList)[i];
		if (inst.HasNullStartTime() || inst.HasNullEndTime()) continue;
		if (inst.StartTime().Int() <= now_min &&
				inst.EndTime().Int() >= now_min) {
			current->AppendL(inst);
		} else if (inst.StartTime().Int() < now_min &&
				inst.EndTime().Int() >= now_min-15 ) {
			prev->AppendL(inst);
		} else if ( inst.EndTime().Int() > now_min &&
				inst.StartTime().Int() <= now_min+15 ) {
			next->AppendL(inst);
		}
	}

	MakeCalendarEvent(current_cal.iPrevious, prev.get(), iModel);
	MakeCalendarEvent(current_cal.iCurrent, current.get(), iModel);
	MakeCalendarEvent(current_cal.iNext, next.get(), iModel);

	if (current_cal==iValue) return;

	iValue=current_cal;
	post_new_value(&iValue);

}

void CLogCalendarImpl::ReadEntriesL(TBool aForce)
{
	CALLSTACKITEM_N(_CL("CLogCalendarImpl"), _CL("ReadEntriesL"));
	if (iState!=EOpened) return;
	if (iModelState!=CAgnEntryModel::EOk) return;

	TTime now; now.HomeTime();

	TDateTime dt=now.DateTime();
	dt.SetHour(0); dt.SetMinute(0); dt.SetSecond(0);
	if ( !aForce && TTime(dt) < TTime(iDayListDate) ) return;
	delete iDayList; iDayList=0;
	iDayList=CAgnDayDateTimeInstanceList::NewL(now);

	TAgnFilter filter;
	filter.SetIncludeEvents(EFalse);
	filter.SetIncludeTodos(EFalse);

	iDayListDate=now.DateTime();
	iModel->PopulateDayDateTimeInstanceListL(iDayList, filter, iDayListDate);
}

void CLogCalendarImpl::expired(CBase* Source)
{
	ReadEntriesL();
	ListCurrentL();
}

CLogCalendarImpl::~CLogCalendarImpl()
{
	CALLSTACKITEM_N(_CL("CLogCalendarImpl"), _CL("~CLogCalendarImpl"));
	delete iModel;

	if (iAgnServ) iAgnServ->Close();
	delete iAgnServ;

	delete iDayList;
	delete iTimeOut;
}