#include "lookup_notif.h"
#include "symbian_auto_ptr.h"
#include "blackboardclientsession.h"
#include "contextcommon.h"
#include "blackboard_cs.h"
#include <s32mem.h>
#include "bbutil.h"

class CLookupHandlerImpl : public CLookupHandler {
	CLookupHandlerImpl(MOutgoingLookup& aOutgoing);
	void ConstructL();
	~CLookupHandlerImpl();

	virtual void IncomingLookupL(const TDesC& aTo, const MBBData* aData);
	virtual void GetNextLookupL(); // will do an async callback with the next one
	virtual void AckLookupL(TUint aId);

	void CheckedRunL();
	void ConnectL();
	void SetFilterL();
	void WaitForNotify();
	void WriteL(const TDesC& aBuf);
	void DoCancel();

	RBBClient	iBBClient;

	HBufC8*		iSerializedData;
	TPtr8		iP;

	TFullArgs	iFullArgs;
	TInt		iAsyncErrorCount;
	MBBDataFactory*	iFactory;
	MOutgoingLookup& iOutgoing;

	friend CLookupHandler;
	friend auto_ptr<CLookupHandlerImpl>;
};

_LIT(KLookupHandler, "CLookupHandler");

CLookupHandler::CLookupHandler() : CCheckedActive(EPriorityStandard, KLookupHandler) { }

CLookupHandler* CLookupHandler::NewL(MOutgoingLookup& aOutgoing)
{
	auto_ptr<CLookupHandlerImpl> ret(new (ELeave) CLookupHandlerImpl(aOutgoing));
	ret->ConstructL();
	return ret.release();
}


CLookupHandlerImpl::CLookupHandlerImpl(MOutgoingLookup& aOutgoing) : 
	iP(0, 0), iOutgoing(aOutgoing) { }

void CLookupHandlerImpl::ConstructL()
{
	iSerializedData=HBufC8::NewL(2048);

	iFactory=CBBDataFactory::NewL();

	ConnectL();
	SetFilterL();
	
	CActiveScheduler::Add(this);
}

CLookupHandlerImpl::~CLookupHandlerImpl()
{
	CALLSTACKITEM_N(_CL("CLookupHandlerImpl"), _CL("~CLookupHandlerImpl"));

	Cancel();
	iBBClient.Close();
	delete iSerializedData;
	delete iFactory;
}

void CLookupHandlerImpl::IncomingLookupL(const TDesC& aTo, const MBBData* aData)
{
	User::Leave(KErrNotSupported);
}
void CLookupHandlerImpl::GetNextLookupL()
{
	WaitForNotify();
}

void CLookupHandlerImpl::AckLookupL(TUint aId)
{
	TRequestStatus s;
	iBBClient.Delete(aId, s);
	User::WaitForRequest(s);
	if (s==KErrNotFound) return;
	User::LeaveIfError(s.Int());
}

void CLookupHandlerImpl::ConnectL()
{
	CALLSTACKITEM_N(_CL("CLookupHandlerImpl"), _CL("ConnectL"));

	Cancel();
	TInt errorcount=0;
	TInt err=KErrNone;
	while (errorcount<5) {
		iBBClient.Close();
		err=iBBClient.Connect();
		if (err==KErrNone) return;
		errorcount++;
	}
	User::Leave(err);
}

void CLookupHandlerImpl::SetFilterL()
{
	CALLSTACKITEM_N(_CL("CLookupHandlerImpl"), _CL("SetFilterL"));

	TInt errorcount=0, err=KErrNone;
	while (errorcount<5) {
		TRequestStatus s;
		iBBClient.AddNotificationL(KOutgoingLookupTuple, ETrue, EBBPriorityNormal, s);
		User::WaitForRequest(s);
		err=s.Int();
		if (err==KErrNone) return;
		ConnectL();
		errorcount++;
	}
	User::Leave(err);
}

void CLookupHandlerImpl::WaitForNotify()
{
	CALLSTACKITEM_N(_CL("CLookupHandlerImpl"), _CL("WaitForNotify"));

	if (IsActive()) return;

	iSerializedData->Des().Zero();
	iP.Set(iSerializedData->Des());
	iBBClient.WaitForNotify(iFullArgs, iP, iStatus);
	SetActive();
}

void CLookupHandlerImpl::DoCancel()
{
	CALLSTACKITEM_N(_CL("CLookupHandlerImpl"), _CL("DoCancel"));

	iBBClient.CancelNotify();
}

_LIT(KValue, "value");

void CLookupHandlerImpl::CheckedRunL()
{
	CALLSTACKITEM_N(_CL("CLookupHandlerImpl"), _CL("CheckedRunL"));

	{
		if (iStatus.Int()!=KErrNone) {
			if (iStatus.Int()==EBufferTooSmall) {
				iSerializedData->Des().Zero();
				iSerializedData=iSerializedData->ReAllocL(iSerializedData->Des().MaxLength()*2);
				iAsyncErrorCount=0;
				WaitForNotify();
				return;
			}
			if (iAsyncErrorCount>5) User::Leave(iStatus.Int());
			ConnectL();
			SetFilterL();
			WaitForNotify();
			return;
		}
	}

	MBBData* d=0;
	{
		RDesReadStream rs(*iSerializedData);
		CleanupClosePushL(rs);
		TTypeName read_type=TTypeName::IdFromStreamL(rs);
		{
			d=iFactory->CreateBBDataL(read_type, KValue, iFactory);
			CleanupPushBBDataL(d);
		}
		{
			d->InternalizeL(rs);
		}
		CleanupStack::Pop();
		CleanupStack::PopAndDestroy();
	}

	{
		bb_auto_ptr<MBBData> p(d);
		iOutgoing.OutgoingLookupL(iFullArgs.iId,
			iFullArgs.iSubName, d);
	}

}
