#include "break.h"
#include <aknutils.h>
#include <txtrich.h>
#include <errorinfo.h>
#include "cu_errorui.h"
#include <eikrted.h>
#include <app_context.h>
#include <symbian_auto_ptr.h>
#include <bautils.h>
#include <aknviewappui.h>
#include <contextui.rsg>

void AppendFormatted(CRichText& aInto, const TDesC& aText, const TRgb& aColor)
{
	TInt pos=aInto.DocumentLength();
	TFontSpec fontspec = LatinPlain12()->FontSpecInTwips();
	TCharFormat charFormat( fontspec.iTypeface.iName, fontspec.iHeight );
	TCharFormatMask charFormatMask;
	charFormat.iFontPresentation.iTextColor=aColor;
	charFormatMask.SetAttrib(EAttColor);

	aInto.InsertL(pos, aText);
	aInto.ApplyCharFormatL(charFormat, charFormatMask, pos, aText.Length());
}

void UserFriendlyIntoRichText(CRichText& aInto, const MErrorInfo& aFrom)
{
	TBuf<14> pre;
	TRgb color;
	switch (aFrom.Severity()) {
	case EInfo:
		pre=_L("[Info] ");
		color=KRgbBlue;
		break;
	case EWarning:
		pre=_L("[Warning] ");
		color=KRgbYellow;
		break;
	case EError:
		pre=_L("[Error] ");
		color=KRgbRed;
		break;
	case ECorrupt:
		pre=_L("[Critical] ");
		color=KRgbDarkRed;
		break;
	}

	AppendFormatted(aInto, pre, color);
	AppendFormatted(aInto, aFrom.UserMessage(), color);
	aInto.InsertL(aInto.DocumentLength(), CEditableText::EParagraphDelimiter);

	switch (aFrom.ErrorType()) {
	case EBug:
		AppendFormatted(aInto, _L("Contact support."), color);
		break;
	case EInputData:
		AppendFormatted(aInto, _L("Please correct the input."), color);
		break;
	case ETemporary:
		AppendFormatted(aInto, _L("Please try again."), color);
		break;
	case ELocalEnvironment:
		AppendFormatted(aInto, _L("There is a problem with the setup on the phone."), color);
		break;
	case ERemote:
		AppendFormatted(aInto, _L("There is a problem with the remote system."), color);
		break;
	}
	aInto.InsertL(aInto.DocumentLength(), CEditableText::EParagraphDelimiter);

}

void TechnicalIntoRichText(CRichText& aInto, const MErrorInfo& aFrom)
{
	AppendFormatted(aInto, _L("Technical information:"), KRgbBlack);
	aInto.InsertL(aInto.DocumentLength(), CEditableText::EParagraphDelimiter);
	AppendFormatted(aInto, aFrom.TechMessage(), KRgbBlack);
	aInto.InsertL(aInto.DocumentLength(), CEditableText::EParagraphDelimiter);

	AppendFormatted(aInto, _L("Error code: "), KRgbBlack);
	TBuf<24> code; code.AppendNum( aFrom.ErrorCode().iUid, EHex );
	code.Append(_L(" "));
	code.AppendNum( aFrom.ErrorCode().iCode );
	AppendFormatted(aInto, code, KRgbBlack);
	aInto.InsertL(aInto.DocumentLength(), CEditableText::EParagraphDelimiter);

	AppendFormatted(aInto, _L("Stack trace: "), KRgbBlack);
	aInto.InsertL(aInto.DocumentLength(), CEditableText::EParagraphDelimiter);
	AppendFormatted(aInto, aFrom.StackTrace(), KRgbBlack);
	aInto.InsertL(aInto.DocumentLength(), CEditableText::EParagraphDelimiter);
}

EXPORT_C void ErrorIntoRichTextL(CRichText& aInto, const MErrorInfo& aFrom)
{
	TInt	nodecount=0;

	const MErrorInfo* list_i=&aFrom;
	while (list_i) {
		const MErrorInfo* nest_i=list_i;
		while (nest_i) {
			nodecount++;
			UserFriendlyIntoRichText(aInto, *nest_i);
			nest_i=nest_i->InnerError();
		}
		list_i=list_i->NextError();
	}
	if (nodecount==1) {
		TechnicalIntoRichText(aInto, aFrom);
	} else {
		list_i=&aFrom;
		while (list_i) {
			const MErrorInfo* nest_i=list_i;
			while (nest_i) {
				nodecount++;
				UserFriendlyIntoRichText(aInto, *nest_i);
				TechnicalIntoRichText(aInto, *nest_i);
				nest_i=nest_i->InnerError();
			}
			list_i=list_i->NextError();
		}
	}
}

class CErrorInfoViewImpl : public CErrorInfoView, public MContextBase {
private:
	CErrorInfoViewImpl();
	void ConstructL();

	virtual void ShowWithData(const MErrorInfo& aInfo);
	
	TUid Id() const;

        void HandleCommandL(TInt aCommand);
        void DoActivateL(const TVwsViewId& aPrevViewId,
		TUid aCustomMessageId,
		const TDesC8& aCustomMessage);
        void DoDeactivate();

	friend class CErrorInfoView;

	CEikRichTextEditor* iContainer;
	MErrorInfo*		iErrorInfo;
	TVwsViewId	iPrevView;
	TInt		iResource;
public:
	virtual ~CErrorInfoViewImpl();
};

EXPORT_C CErrorInfoView* CErrorInfoView::NewL()
{
	CALLSTACKITEM_N(_CL("CErrorInfoView"), _CL("NewL"));

	auto_ptr<CErrorInfoViewImpl> ret(new (ELeave) CErrorInfoViewImpl);
	ret->ConstructL();
	return ret.release();
}
CErrorInfoViewImpl::CErrorInfoViewImpl() { }

void CErrorInfoViewImpl::ConstructL()
{
	CALLSTACKITEM_N(_CL("CErrorInfoViewImpl"), _CL("ConstructL"));
#ifndef __WINS__
	TFileName resfile=_L("c:\\System\\data\\contextui.rsc");
	if (! BaflUtils::FileExists(iEikonEnv->FsSession(), resfile) ) {
		resfile.Replace(0, 1, _L("e"));
	}
#else
	TFileName resfile=_L("z:\\System\\data\\contextui.rsc");
	BaflUtils::NearestLanguageFile(iEikonEnv->FsSession(), resfile); //for localization
#endif
	CC_TRAPD(err, iResource=iEikonEnv->AddResourceFileL(resfile));
	User::LeaveIfError(err);

	BaseConstructL( R_ERRORINFO_VIEW );
}

TUid CErrorInfoViewImpl::Id() const
{
	CALLSTACKITEM_N(_CL("CErrorInfoViewImpl"), _CL("Id"));

	return KErrorInfoView;
}

void CErrorInfoViewImpl::HandleCommandL(TInt aCommand)
{
	CALLSTACKITEM_N(_CL("CErrorInfoViewImpl"), _CL("HandleCommandL"));
	ActivateViewL(iPrevView);
}

void CErrorInfoViewImpl::DoActivateL(const TVwsViewId& aPrevViewId,
	TUid /*aCustomMessageId*/,
	const TDesC8& /*aCustomMessage*/)
{
	CALLSTACKITEM_N(_CL("CErrorInfoViewImpl"), _CL("DoActivateL"));

	iPrevView=aPrevViewId;
	if (!iContainer) {
		iContainer=new (ELeave) CEikRichTextEditor();
		iContainer->ConstructL( 0, 10, 2000, 0 );
		iContainer->SetRect( ClientRect() );
		iContainer->SetMopParent(this);
		iContainer->CreatePreAllocatedScrollBarFrameL();

		if (iErrorInfo) {
			ErrorIntoRichTextL( *iContainer->RichText(), *iErrorInfo );
			delete iErrorInfo; iErrorInfo = 0;
		}
		AppUi()->AddToStackL( *this, iContainer );
        } 
}

void CErrorInfoViewImpl::DoDeactivate()
{
	CALLSTACKITEM_N(_CL("CErrorInfoViewImpl"), _CL("DoDeactivate"));

	if ( iContainer )
        {
		AppUi()->RemoveFromViewStack( *this, iContainer );
        }
	
	delete iContainer; iContainer = 0;
	delete iErrorInfo; iErrorInfo = 0;
}

CErrorInfoViewImpl::~CErrorInfoViewImpl()
{
	CALLSTACKITEM_N(_CL("CErrorInfoViewImpl"), _CL("~CErrorInfoViewImpl"));

	if (iResource) iEikonEnv->DeleteResourceFile(iResource);
	delete iContainer;
	delete iErrorInfo;
}

void CErrorInfoViewImpl::ShowWithData(const MErrorInfo& aInfo)
{
	delete iErrorInfo; iErrorInfo=0;
	iErrorInfo=aInfo.CreateCopyL();
	AppUi()->ActivateLocalViewL(KErrorInfoView);
}
